from fs.osfs import OSFS
from .model import File


def touch(f : File):
    with OSFS('/') as fs:
        fs.create(f.path.relative)


def mv(f : File, path_new, overwrite=True):
    with OSFS('/') as fs:
        fs.move(f.path.relative, path_new.relative, overwrite)


def cp(f : File, path_new, overwrite=True):
    with OSFS('/') as fs:
        fs.copy(f.path.relative, path_new.relative, overwrite)


def rm(f : File):
    with OSFS('/') as fs:
        fs.remove(f.path.relative)



def add_api(api, root, name='fs'):
    from .path.web import add_api as add_path_api
    from .file.web import add_api as add_file_api
    from .directory.web import add_api as add_dir_api
    root_fs = root + '/{0}'.format(name)
    add_path_api(api, root_fs)
    add_file_api(api, root_fs)
    add_dir_api(api, root_fs)


from fs.osfs import OSFS
from ..path.model import Path
from ..file.model import File
import pathlib
import yaml


class Directory:
    yaml_tag = '!directory'

    def __init__(self, path, load_depth=0):
        if isinstance(path,Path):
            self.path = path
        else:
            self.path = Path(path)
        if load_depth > 0:
            self.load(load_depth)
        else:
            self.children = None
            

    def load(self, depth):
        self.children = None
        if depth < 0:
            return
        if not self.exists:
            raise DirectoryNotFoundError(self.path.abs)
        if depth == 0:
            return
        self.children = []
        with OSFS('/') as fs:
            for f in fs.listdir(self.path.relative):
                fp = self.path / f
                if fs.isdir(fp.relative):
                    self.children.append(Directory(fp, depth - 1))
                else:
                    self.children.append(File(fp, 0))

    @property
    def is_dir(self):
        return pathlib.Path(self.path.abs).is_dir()

    @property
    def exists(self):
        with OSFS('/') as fs:
            return fs.exists(self.path.relative) and fs.isdir(self.path.relative)

    def ensure(self):
        if not self.exists:
            with OSFS('/') as fs:
                fs.makedirs(self.path.rel)

    def to_serializable(self):
        return {'path': self.path.abs,
                'name': self.path.name,
                'is_dir': True,
                'children': [c.to_serializable() for c in self.children] if self.children is not None else None
                }

    def __str__(self):
        import json
        return json.dumps(self.to_serializable(), sort_keys=True, separators=(',', ':'), indent=4)



from fs.osfs import OSFS
from .model import Directory


def mkdir(d:Directory, recreate=True):
    with OSFS('/') as fs:
        fs.makedirs(d.path.relative, recreate=recreate)


def mv(d:Directory, path_new):
    with OSFS('/') as fs:
        fs.movedir(d.path.relative, path_new.relative)


def cp(d:Directory, path_new, create=True):
    with OSFS('/') as fs:
        fs.copydir(d.path.relative, path_new.relative, create)


def rm(d:Directory):
    with OSFS('/') as fs:
        fs.removetree(d.path.relative)


from ..path import Path

class FileSystem:
    def __init__(self, fs_or_path):
        from .config import c as config
        if fs_or_path is None:
            self.fs = config.default_fs('/')
            self.need_close = True
        elif isinstance(fs_or_path, str):
            self.fs = config.default_fs(Path(fs_or_path).abs)
            self.need_close = True
        else:
            self.fs = fs_or_path
            self.need_close = False

    def __enter__(self):
        return self.fs

    def __exit__(self, type, value, trackback):
        if self.need_close:
            self.fs.close()
            



from setuptools import setup, find_packages
setup(
    name='jfs',
    version='0.1.0',
    description='File Processing Library.',
    url='https://github.com/twj2417/filesystem',
    author='Weijie Tao',
    author_email='twj2417@gmail.com',
    license='MIT',
    namespace_packages=['jfs'],
    packages=find_packages('src/python'),
    package_dir={'': 'src/python'},
    install_requires=[
        'click',
        'rx',
        'pathlib',
        'fs',
    ],
    scripts=[],
    zip_safe=False
)



from .directory import Directory
from .directory import mv as dmv
from .directory import rm as drm
from .directory import cp as dcp
from .file import File
from .file import mv as fmv
from .file import rm as frm
from .file import cp as fcp


def mv(sor, tar, overwrite=True):
    if isinstance(sor, Directory):
        dmv(sor, tar)
    else:
        fmv(sor, tar, overwrite)


def cp(sor, tar, overwrite=True):
    if isinstance(sor, Directory):
        dcp(sor, tar, overwrite)
    else:
        fcp(sor, tar, overwrite)


def rm(sor):
    if isinstance(sor, Directory):
        drm(sor)
    else:
        frm(sor)


from flask import request
from flask_restful import Resource
from .model import File
from . import service
from ..exceptions import http_error


class FileResource(Resource):
    def get(self, path):
        try:
            is_load = request.args.get('is_load')
            if is_load is None or is_load is False:
                load_depth = 0
            else:
                load_depth = 1
            return File(path, load_depth).to_serializable(), 200
        except Exception as e:
            return http_error(e)

    def put(self, path):
        try:
            method = request.form.get('method')
            if method == 'cp':
                service.cp(File(path), request.form.get('target'))
            elif method == 'mv':
                service.mv(File(path), request.form.get('target'))
            elif method == 'modify':
                File(path).save(request.form.get('data'))
            return {'status': 'OK'}, 200
        except Exception as e:
            return http_error(e)

    def delete(self, path):
        try:
            service.rm(File(path))
            return {'status': 'OK'}, 200
        except Exception as e:
            return http_error(e)


class FilesResource(Resource):
    def post(self):
        try:
            service.touch(File(request.form.get('path')))
            return {'status': 'OK'}, 201
        except Exception as e:
            return http_error(e)


def add_api(api, root, name='file'):
    api.add_resource(FileResource, root + '/{0}/<path>'.format(name))
    api.add_resource(FilesResource, root + '/{0}s/<path>'.format(name))


"""
Abstract of path object.
Features:
    An unified model for posix/windows/url/url_quoted paths.
"""

import pathlib 
import fs.path as fp
from dxl.data import DataClass
from dxl.data.maybe import Maybe
from urllib.parse import quote_plus,unquote_plus,urlparse

def _unified_protocol(p:str) ->str:
    if not p.endswith('@'):
        return p + '@'
    return p


def _divide_protocol(p: str) -> (str, str):
    n = p.find('@')
    if n >= len(p) - 1:
        return p[:], ''
    if n == -1:
        return None, p
    else:
        return p[:n + 1], p[n + 1:]

def _decode_url(p: str):
        return unquote_plus(unquote_plus(p))

def _unified_path(p)-> str:
        if isinstance(p, Path):
            return p._p
        if not isinstance(p, str):
            raise TypeError(
                "{} is not convertable to {}.".format(type(p), __class__))
        protocol, raw_path = _divide_protocol(p)
        raw_path = _decode_url(raw_path)
        raw_path = fp.normpath(raw_path)
        if protocol is None:
            return raw_path
        return protocol + raw_path

class Path:
    def __init__(self,path,protocol=None):
        if protocol is not None:
            path = _unified_protocol(protocol) + path
        self._p = _unified_path(path)
               
    @property
    def protocol(self) -> str:
        return _divide_protocol(self._p)[0]

    @property
    def raw(self) -> str:
        return _divide_protocol(self._p)[1]

    def raw_path(self) -> 'Path':
        """
        Returns raw path object, i.e. no protocol.
        """
        return Path(self.raw, None)

    @property
    def route(self) -> str:
        "Alias of self.raw_path"
        return self.raw

    @property
    def _r(self):
        return pathlib.Path(self.route)

    @property
    def abs(self) -> str:
        #return fp.abspath(self.route)
        if fp.isabs(self.route):
            return self.route
        else:
            from fs.osfs import OSFS
            with OSFS('.') as fs:
                return fs.getsyspath(self.route)

    def absolute(self) ->'Path':
        return Path(self.abs,self.protocol)

    @property
    def relative(self) ->str:
        return fp.relpath(self.route)

    @property
    def isroot(self) -> bool:
        return self.route == '/'
    
    @property
    def isabs(self):
        return fp.isabs(self.route)

    @property
    def name(self) ->str:
        return fp.basename(self.route)

    @property
    def father(self) ->str:
        return fp.dirname(self.abs)

    def father_path(self) ->'Path':
        return Path(self.father,self.protocol)

    def parts(self):
        return self._r.parts

    def is_exist(self):
        return self._r.exists()

    def join(self,name):
        return Path(fp.combine(self.route,name),self.protocol)

    def __truediv__(self, name):
        return self.join(name)

    def __add__(self, suffix):
        return Path(self.route + suffix, self.protocol)

    def __str__(self):
        return self._p

    def __hash__(self):
        return hash(self._p)

    def __eq__(self, path):
        return self._p == Path(path)._p

    @property
    def suffix(self):
        return self._r.suffix

    @property
    def suffixes(self):
        return self._r.suffixes

    


class NonePathError(TypeError):
    def __init__(self):
        super(__class__, self).__init__("Path can not be None type.")


class NotUrlPathError(ValueError):
    def __init__(self, path):
        super(__class__, self).__init__(
            "Expected an url quoted path, got {}.".format(path))


class IsUrlPathError(ValueError):
    def __init__(self, path):
        super(__class__, self).__init__(
            "Expected an not url quoted path, got {}.".format(path)
        )


class ConflictUrlSpecError(ValueError):
    def __init__(self, us1, us2):
        super(__class__, self).__init__(
            "Conflict UrlSpecs: {0} and {1}".format(us1, us2))


class NotValidPathError(ValueError):
    def __init__(self, path):
        super(__class__, self).__init__(
            "Got an invalid path: {}.".format(path)
        )


class DirectoryNotFoundError(ValueError):
    def __init__(self, path):
        super(__class__, self).__init__(
            "Directory {0} not found.".format(path))


def http_error(e):
    def emsg(text):
        return {'status': 'ERROR', 'message': text}
    if isinstance(e, FileNotFoundError):
        return emsg("File {} not found.".format(e)), 404
    elif isinstance(e, DirectoryNotFoundError):
        return emsg("Directory {} not found.".format(e)), 404
    elif isinstance(e, NotADirectoryError):
        return emsg("{} is not a directory.".format(e)), 409
    elif isinstance(e, NonePathError):
        return emsg('Path can not be None.'), 400
    elif isinstance(e, FileExistsError):
        return emsg('File {} exists.'.format(e)), 409
    else:
        return emsg('Unknown error: {}'.format(e)), 500